// Type definitions for binary-tools 1.11
// Project: https://rochars.com/binary-tools
// Definitions by: Rafael da Silva Rocha <rocha.rafaelsilva@gmail.com>

// packing/unpacking

/**
 * Pack values to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!Array<number|string|bigint>} values The values to pack.
 * @param {!Array<number>} buffer The byte buffer to write on.
 * @param {number=} [index=0] The buffer index to start writing.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write.
 * @throws {Error} On unsupported type.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If 'values' contains invalid values for the types.
 */
export function packTo(
  theType: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  buffer: ArrayLike<number>,
  index?: number,
  clamp?: boolean): number;

/**
 * Unpack values from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {!Array<number|string|bigint|boolean>} output The output Array.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpackTo(
  format: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number|string|bigint>,
  index?: number,
  safe?: boolean): void;

/**
 * Pack values as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!Array<number|string|bigint>} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not a number.
 * @throws {TypeError} If 'values' is not a int and type is int.
 */
export function pack(
  theType: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  clamp?: boolean): Array<number>;

/**
 * Unpack values from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!Array<number|string|bigint|boolean>} The unpacked values.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpack(
  theType: string,
  buffer: ArrayLike<number>,
  index?: number,
  safe?: boolean): Array<number|string|bigint>;

// reading/writing

/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<number|string|bigint>} values The values to write.
 * @param {!Array<number>} buffer The buffer to write on.
 * @param {number=} [index=0] The buffer index to start writing.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If 'values' contains invalid values for the types.
 */
export function writeTo(
  theType: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  buffer: ArrayLike<number>,
  index?: number,
  clamp?: boolean): number;

/**
 * Read a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer.
 * @param {!Array<number|string|bigint|boolean>} output The output array.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function readTo(
  theType: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number|string|bigint>,
  start?: number,
  end?: number,
  safe?: boolean): void;

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!Array<number|string|bigint>} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not a array of numbers.
 * @throws {TypeError} If 'values' is not a array of ints and type is int.
 */
export function write(
  theType: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  clamp?:boolean): Array<number>;

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!Array<number|string|bigint|boolean>}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function read(
  theType: string,
  buffer: ArrayLike<number>,
  start?: number,
  end?: number,
  safe?: boolean): Array<number|string|bigint>;

// tools

/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!Array<number>} buffer The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 * @function endianness
 * @memberOf module:binary-tools
 */
export function endianness(
  buffer: ArrayLike<number>,
  offset: number,
  start?: number,
  end?: number): void;

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @param {boolean=} [includePads=false] True to include pads in the count.
 * @return {number}
 * @throws {Error} On unsupported type.
 * @function calcSize
 * @memberOf module:binary-tools
 */
export function calcSize(
  format: string,
  includePads?: boolean): number;

/**
 * Returns how many bytes are needed to serialize a UTF-8 string.
 * @param {string} str The string to pack.
 * @return {number} The number of bytes needed for the string.
 */
export function utf8BufferSize(
  str: string): number

/**
 * Returns how many bytes are needed to serialize a UTF-16 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 2).
 * @return {number} The number of bytes needed for the string.
 */
export function utf16BufferSize(
  str: string,
  forceBOM?: boolean): number

/**
 * Returns how many bytes are needed to serialize a UTF-32 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 4).
 * @return {number} The number of bytes needed for the string.
 */
export function utf32BufferSize(
  str: string,
  forceBOM?: boolean): number

/**
 * Format a byte array as a string of hex numbers.
 * @param {!Array<number>} bytes The bytes.
 * @return {string}
 * @throws {RangeError} If bytes contains invalid values.
 */
export function bytesToHex(
  bytes: ArrayLike<number>): string

/**
 * Convert a hex string to a byte array.
 * @param {string} hexStr The hex string.
 * @return {!Array<number>}
 * @throws {RangeError} If string contains chars outside the 0..f range.
 * @throws {Error} If string length is not even.
 */
export function hexToBytes(
  hexStr: string): Array<number>

// types

/**
 * A binary-tools datatype definition object.
 * @typedef {Object.<string, number|boolean>} DataType
 * @property {number} bits The number of bits used by the type.
 * @property {boolean} signed False for unsigned, true for signed.
 * @property {boolean} fp True for floating point numbers only.
 * @property {boolean} c True for single-byte character encodings.
 * @property {boolean} b Alternative type. Used only by types g and N.
 * @property {boolean} v Variable length. Only true for string types.
 * @property {boolean} u False for all except UTF strings.
 * @memberOf module:binary-tools
 */

/**
 * The standard binary-tools datatype definitions.
 * Types can be created by adding new properties named as
 * a single character to this object in the format:
 * { bits: number, signed: boolean, fp: boolean }
 * Numbers and the characters \@~=<>! are reserved, and also
 * all characters used for pre-defined types (xcCsSuUVNbB?hHegtTlLfjJkKdqQAYOM).
 * @type {Object.<string, Object>}
 * @property {DataType} x Pad byte
 * @property {DataType} c ASCII character
 * @property {DataType} C Extended ASCII (ISO-8859-1) character
 * @property {DataType} s ASCII string
 * @property {DataType} S Extended ASCII (ISO-8859-1) string
 * @property {DataType} u UTF-8 string
 * @property {DataType} U UTF-16 string
 * @property {DataType} V UTF-32 string
 * @property {DataType} N Unsiged 4-bit integers
 * @property {DataType} b Signed 8-bit integers
 * @property {DataType} B Unsigned 8-bit integers
 * @property {DataType} ? Boolean
 * @property {DataType} h Signed 16-bit integers
 * @property {DataType} H Unsigned 16-bit integers
 * @property {DataType} e 16-bit Half-precision floating-point numbers
 * @property {DataType} g 16-bit Brain floating point numbers
 * @property {DataType} t Signed 24-bit integers
 * @property {DataType} T Unsigned 24-bit integers
 * @property {DataType} i Signed 32-bit integers
 * @property {DataType} I Unsigned 32-bit integers
 * @property {DataType} l Signed 32-bit integers
 * @property {DataType} L Unsigned 32-bit integers
 * @property {DataType} f 32-bit Single-precision floating-point numbers
 * @property {DataType} j Signed 40-bit integers
 * @property {DataType} J Unsigned 40-bit integers
 * @property {DataType} k Signed 48-bit integers
 * @property {DataType} K Unsigned 48-bit integers
 * @property {DataType} d 64-bit Double-precision floating-point numbers
 * @property {DataType} q signed 64-bit integers
 * @property {DataType} Q Unsigned 64-bit integers
 * @property {DataType} A Unsigned 128-bit integers
 * @property {DataType} Y Unsigned 256-bit integers
 * @property {DataType} O Unsigned 512-bit integers
 * @property {DataType} M Unsigned 1024-bit integers
 * @memberOf module:binary-tools
 */
export const types;
